/*
 * Kangaroo Punch MultiPlayer Game Server Mark II
 * Copyright (C) 2020-2021  Scott Duensing
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */


#ifndef VESA20_H
#define VESA20_H


#include "os.h"
#include "color.h"


uint8_t   vbeDisplayDepthBitsGet(void);
uint8_t   vbeDisplayDepthBytesGet(void);
uint16_t  vbeDisplayHeightGet(void);
uint16_t  vbeDisplayWidthGet(void);
ColorT    vbeColorMake(uint8_t red, uint8_t green, uint8_t blue);
void      vbePresent(void);
int16_t   vbeInfoShow(void);
int16_t   vbeShutdown(void);
uint8_t   vbeStartup(uint16_t xRes, uint16_t yRes, uint8_t bpp);
void      vbeVBlankWait(void);


#endif  // VESA20_H
